#+++
#  title:	Ctabdiff.awk
#  abstract:	Generate character table diff.
#  author:	T.R.Hageman, Groningen, The Netherlands.
#  created:	march 1992
#  modified:	(see Log at end)
#  description:
#
#  	usage: awk -f Ctabdiff.awk [debug=1] <base.tab> <target.tab>
#
#	Compare two sets of tables (as generated by Mkctable.awk).
#
#	A matching entry is replaced with 0; if an entry differs the value
#	from <target.tab> is written.
#
#	Difference tables are written to stdout, with identical heads and
#	tails removed.  The resulting start index of the diff tables is output
#	as `enum { ID(offset) = ... };'
#
#	The Accent table and character-set name are taken from <target.tab>
#	as-is.
#
#	Identifiers are output as ID(name), so the preprocessor macro ID()
#	can (and should) be defined before inclusion to turn these into
#	valid C identifiers.
#
#	See Mkctable.awk for this to make any sense.
#
#	For an example of how to use this, see the SELECTABLE conditional
#	code in the JOVE source files msdos.c and ctype.c
#---

BEGIN {
	split("$Id: Ctabdiff.awk,v 1.7 1993/11/04 21:39:21 tom Exp tom $", Id)
	if (i = index(Id[2], ","))
		ScriptName = substr(Id[2], 1, i - 1)
	else if ((ScriptName = Id[2]) == "")
		ScriptName = "ctabdiff.awk"
	done = 0
	intable = 0
	debug = 0
	lno = 0
}

Filename != FILENAME {
	NFiles++;
	if (NFiles > 2) {
		print "/* too many files, bailing out */"
		exit
	}
	Filenames[NFiles] = FILENAME
	Filename = FILENAME
	FileStart = NR - 1
	name = ""
}

/^public Const.*\[CTABLESIZE\] = {/ {		# byte table
	i = substr($4, 1, index($4, "[") - 1)
	intable = 1
	cc = $(NF - 4)
	if (i == "UpcTable") {
		intable = 2			# ignore
	}
	else if ((tt = tablenum[name = i]) == "") {
		tablename[++Ntables] = name
		tablenum[name] = Ntables
		start[name] = cc
		if (debug) printf("/* New table \"%s\" (%d), [%d-%d) */\n", \
				  name, Ntables, cc, $(NF - 2))
	}
	else {
		if (cc < start[name])
			start[name] = cc;
		if (debug) printf("/* table \"%s\" found */\n", name)
	}
	prefixlength[name] = lno
	next
}

/^}/ && intable != 0 {
	if (intable == 1) {
		if (cc > end[name])
			end[name] = cc
	}
	intable = 0
	next
}

intable == 1 {
	if ($0 !~ /^\t.*/) next

	if (NFiles == 1) {
		for (i = 1;  i <= NF;  i++) {
			table[cc "," name] = $i
			cc++
		}
	}
	else {
		for (i = 1;  i <= NF;  i++) {
			if (table[cc "," name] != $i)
				table[cc "," name] = $i
			else
				table[cc "," name] = "0,"
			cc++
		}
	}
	next
}

NFiles == 1 || intable != 0 { next }

name == "" && headerSkipped == "" {
	if ($0 ~ /.*\*\//) headerSkipped = 1
	next
}

# second file only

/^private.*charset_name\[\]/ {
	$4 = "ID(" substr($4, 1, index($4, "[") - 1) ")[]"
	if (debug) printf("/* %s */\n", $4)
}

/^private.*AccentTab\[\]/ {
	$6 = "ID(" substr($6, 1, index($6, "[") - 1) ")[]"
	if (debug) printf("/* %s */\n", $6)
}

{ prefix[lno++] = $0 "" }

END {
	if (NFiles != 2) {
		print "/* usage: awk -f " ScriptName \
		      " [debug=1] <first.tab> <second.tab> */"
		exit
	}
	# calculate constant head and tail of the tables

	headmin = 256; tailmax = 0

	for (tt = 1;  tt <= Ntables; tt++) {
		name = tablename[tt]

		# head
		for (cc = start[name];  cc < end[name];  cc++) {
			if (table[cc "," name] != "0,")
				break
		}
		if (cc < headmin)
			headmin = cc

		# tail
		for (i = end[name]; --i >= cc; ) {
			if (table[i "," name] != "0,")
				break
		}
		if (i >= tailmax && i >= cc)
			tailmax = i + 1

		if (debug) printf("/* table \"%s\"[%d,%d) cc=%d headmin=%d i=%d tailmax=%d */\n", name, start[name], end[name], cc, headmin, i, tailmax) 
	}

	if (tailmax <= headmin) {
		print "/* NO DIFFS */"
		exit
	}

	print "/* This file is generated using the awk script \"" \
	      ScriptName "\" (" Id[3] " " Id[4] ")"
	print " * from \"" Filenames[1] "\" and \"" Filenames[2] "\"\n */\n"
	print "#ifndef ID"
	print "#   define ID(id) id"
	print "#endif\n"
	print "enum { ID(offset) = " headmin " };"

	openformat = "private const char ID(%s)[] = {" \
		     sprintf("	/* [ %d - %d ) */\n%s", \
			     headmin, tailmax, \
			     substr("\t\t\t\t\t\t\t", 8 - headmin % 8))
	closeformat = "};\n"
	if (tailmax % 8)  closeformat = "\n" closeformat

	ll = 0

	for (tt = 1;  tt <= Ntables;  tt++) {
		name = tablename[tt];
		while (ll < prefixlength[name]) print prefix[ll++]
		printf(openformat, name)
		for (i = headmin;  i < tailmax; ) {
			printf("\t%s", table[i "," name])
			if (++i % 8 == 0) {
				print
				if (i % 64 == 0 && i != tailmax)
			}
		}
		printf(closeformat)
	}
	while (ll < lno) print prefix[ll++]
}

#======================================================================
# $Log: Ctabdiff.awk,v $
# Revision 1.7  1993/11/04  21:39:21  tom
# fix bug in handling tables with no diffs (table was not shrunk properly);
# refrain from generating output when all tables are identical.
#
# Revision 1.6  1993/08/04  04:05:10  tom
# generate `enum { ...offset... }' instead of `const int ...offset...',
# so that its value can be used at compile-time.
#
# Revision 1.5  1993/01/26  18:39:07  tom
# add explanatory comment header, and RCS directives.
#
#======================================================================
